/*
 * global_nn_classifier.cpp
 *
 *  Created on: Mar 9, 2012
 *      Author: aitor
 */

#include <pcl/apps/3d_rec_framework/pipeline/impl/global_nn_classifier.hpp>
#include <pcl/apps/3d_rec_framework/utils/metrics.h>

// Instantiation
template class pcl::rec_3d_framework::
    GlobalNNPipeline<flann::L1, pcl::PointXYZ, pcl::VFHSignature308>;
template class pcl::rec_3d_framework::GlobalNNPipeline<
    Metrics::HistIntersectionUnionDistance,
    pcl::PointXYZ,
    pcl::VFHSignature308>;
template class pcl::rec_3d_framework::
    GlobalNNPipeline<flann::L1, pcl::PointXYZ, pcl::ESFSignature640>;

template class pcl::rec_3d_framework::GlobalClassifier<pcl::PointXYZ>;
