#!/bin/bash
../../../build/bin/people_app -numTrees 3 -tree0 ../data/results/forest1/tree_20.txt -tree1 ../data/results/forest2/tree_20.txt -tree2 ../data/results/forest3/tree_20.txt -tree3 ../data/results/forest3/tree_20.txt -mask 0 -FG 0 $1

